#!/usr/bin/python
# cp-run - program for cool plate control
#version 2018-04-02

import time, string, sys, traceback, math, os
import  RPi.GPIO as GPIO
from cp_defs import *
from cp_measure import *

# Initiate history list which contains
# the present and the previous sets of data
# needed for PID control
hist0 = [1]*20
hist1 = [1]*20
pid_int = 0 # accumulator for PID integral term
#set up the starting values
GPIO.setmode(GPIO.BCM) 
GPIO.setup(26, GPIO.OUT) # pump
GPIO.setup(20, GPIO.OUT) # cooler
GPIO.setup(13, GPIO.OUT) # cool
GPIO.setup(6, GPIO.OUT) # heat
#GPIO.setup(xx, GPIO.OUT) # led for program running
#GPIO.setup(xx, GPIO.OUT) # led for program on standby

GPIO.output(13, GPIO.LOW)
GPIO.output(6, GPIO.LOW)
#GPIO.output(xx, GPIO.LOW)
#GPIO.output(xx, GPIO.LOW)
GPIO.output(26, GPIO.HIGH) # starts  the pump
GPIO.output(20, GPIO.HIGH) # starts  the cooler

elapsedTime = 0 # seconds from start
cycleCount = 0
timeOffset= int(time.time() )# seconds since 1970,; relates run time to actual time
timePrev = timeOffset
timeNext = timePrev + 60
# timeOffset is a constant for this run
tgt = 0 # sets arbitrary target temperature
pidTime = 60 # secs PID cycle
heat = 0.0 # secs warming
tError = 0  # accumulator for PID  integral part
#autoClose = 0 # set to 1 if element overheats
tPlate = measure(tgt) # returns a string of date-time t1 t2 ... target
# also writes file t_record.txt
hist0 = hist1 = tPlate.split(' ') # store list for comparison with the next iteration

while 1: # main loop
  try:
    pp = open("params.py", "r")
    cpc = pp.readline()
    p = eval(cpc)
    waitFlag = 0  
    if ((p['stopAt'] == '0') or (elapsedTime > int(p['stopAt']) * 3600)) : # stop sign set from web page 
      # set safe standby
      if (waitFlag == 0):
         GPIO.output(21, GPIO.LOW) # stop heater
         GPIO.output(13, GPIO.LOW) # stop cooli ng
         GPIO.output(20, GPIO.LOW) # stop cooling machine
         GPIO.output(26, GPIO.LOW) # stop pump
         #GPIO.output(xx,GPIO.LOW) # stop action led
         #GPIO.output(xx,GPIO.HIGH) # light standby led
         elapsedTime = 0 # reset timer
         cycleCount = 0
         tError = 0
         # set stop at time to zero
         p['stopAt'] = '0'
         writeparams = open("params.py","w")
         writeparams.write( str(p))
         writeparams.close()
         waitFlag = 1
      timeOffset= int(time.time() )# seconds since 1970,; relates run time to actual time
      timePrev = timeOffset
      timeNext = timePrev + 60
      print ("waiting...")
      #wait for action
      time.sleep(60)
      continue # check for new instructions
    while (int(time.time()) <= timeNext): time.sleep(1)
    timeNext = timeNext + 60
    if ((cycleCount % 5) == 0): # time to re-graph
      os.system("/home/pi/coolplate/programs/cp_graph.sh")
      sys.stdout.flush()
    cycleCount += 1
    #GPIO.output(xx,GPIO.HIGH) # active lamp
    #GPIO.output(xx,GPIO.LOW) # stop standby lamp
    elapsedTime = int(time.time()) - timeOffset # reconnect to absolute time
    
    # calculate fraction into cycle time 
    elapsedSecondsIntoCycle = ((elapsedTime % (float(p['tPeriod'])*3600)) + float(p['startAt'] )* 3600)
    fractionIntoCycle = elapsedSecondsIntoCycle / (float(p['tPeriod'] )* 3600)
    
    cycleType = p['waveform'] 
    
    if (cycleType == "sine"):
       radiansIntoCycle = fractionIntoCycle * 2* math.pi
       tgt = float(p['tAvg']) + float(p['tHalfAmp']) * math.sin(radiansIntoCycle)   
 	    
    if (cycleType == "square"):
       if (fractionIntoCycle <= 0.5):
          tgt = float(p['tAvg']) + float(p['tHalfAmp']) 
       else:  	
          tgt = float(p['tAvg']) - float(p['tHalfAmp']) 
     	
    if (cycleType == "triangle"):
       if (fractionIntoCycle <= 0.25):
          tgt = float(p['tAvg']) + float(p['tHalfAmp']) * fractionIntoCycle * 4 
       elif ((fractionIntoCycle > 0.25) and (fractionIntoCycle <= 0.75)):  	
          tgt = float(p['tAvg']) + float(p['tHalfAmp']) - float(p['tHalfAmp']) * (fractionIntoCycle - 0.25) * 4
       else: # last quarter
          tgt = float(p['tAvg']) - float(p['tHalfAmp'])  + float(p['tHalfAmp']) * (fractionIntoCycle - 0.75) * 4
    	    	
    # measure temperatures
    tPlate = measure(tgt) # returns a string of date-time t1 t2 ... target
    hist1= tPlate.split(' ') # store for comparison with the previous iteration
    print ("\n %s Elapsed s: %s Temperatures: %s  %s Target: %s " % (hist1[0], elapsedTime, hist1[1], hist1[2], hist1[3]))
    tNow = float(hist1[1])
    tPrev = float(hist0[1])
    tgtNow = tgt    
    tgtPrev = float(hist0[3])
    
    # compare actual with target temperatures
    # first reduce previous error, to give fading integral history
    tError = tError*0.7
    tError += float(hist1[1]) - tgt
    heat = cp_pid(tError, tPrev, tNow, tgtPrev, tgtNow, float(p['pid_p']), float(p['pid_i']), float(p['pid_d']))
    print ("heating seconds: %s " % int(heat))
    
    # calculation finished, move present measures to previous
    hist0 = hist1
    
   # heat time over one minute
    if (heat > 55) :  
      GPIO.output(6, GPIO.HIGH)  # heat
      GPIO.output(13, GPIO.LOW)  # no cooling
      continue
    if (heat < -55): 
      GPIO.output(6, GPIO.LOW)
      GPIO.output(13, GPIO.HIGH)
      continue
   # heat within range +/- 60
    if (heat >= 0): 
      GPIO.output(6, GPIO.HIGH)
      GPIO.output(13, GPIO.LOW)
      time.sleep(heat)
      GPIO.output(6, GPIO.LOW)
    if (heat < 0):
      GPIO.output(6, GPIO.LOW)
      GPIO.output(13, GPIO.HIGH)
      time.sleep(abs(heat))
      GPIO.output(6, GPIO.LOW)  
    
  except: # on any un-caught exception
    e = traceback.print_exc()
    print ("\n Error %s \n Restarting" % e)
    GPIO.output(6, GPIO.LOW) # stop heater
    GPIO.output(13, GPIO.LOW) # stop cooling
    GPIO.output(20, GPIO.LOW) # stop cooling machine 
    GPIO.output(26, GPIO.LOW) # stop pump
    #continue # don't give up!
 
    break # should never be reached
  # end of while loop
# end of program
